/* -*- Mode:C++; c-file-style:"gnu"; indent-tabs-mode:nil; -*- */
/*
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation;
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */

#include "ns3/log.h"
#include "ns3/abort.h"
#include "ns3/net-device.h"
#include "ns3/net-device-container.h"
#include "ns3/point-to-point-helper.h"
#include "ns3/ipv4-address-helper.h"
#include "ns3/data-rate.h"
#include "ns3/random-variable-stream.h"

#include "brite-topology-helper.h"

#include <iostream>
#include <fstream>

NS_LOG_COMPONENT_DEFINE ("BriteTopologyHelper");

namespace ns3 {

BriteTopologyHelper::BriteTopologyHelper (std::string confFile, 
                                          std::string seedFile,
                                          std::string newseedFile) : numAs(0), numNodes(0), numEdges(0)
{
  brite::Brite br (confFile, seedFile, newseedFile);
  m_topology = br.GetTopology ();
  BuildBriteNodeInfoList ();
  BuildBriteEdgeInfoList ();
}

BriteTopologyHelper::BriteTopologyHelper (std::string confFile) : numAs(0), numNodes(0), numEdges(0)
{

  //Generate seed file expected by BRITE
  //need unsigned shorts 0-65535
  Ptr<UniformRandomVariable> uv = CreateObject<UniformRandomVariable> ();
  std::ofstream seedFile;
  seedFile.open ("briteSeedFile.txt");
  seedFile << "PLACES " << uv->GetInteger (0, 65535) << " " << uv->GetInteger (0, 65535) << " " << uv->GetInteger (0, 65535) << std::endl;
  seedFile << "CONNECT " << uv->GetInteger (0, 65535) << " " << uv->GetValue () << " " << uv->GetValue () << std::endl;
  seedFile << "EDGE_CONN " << uv->GetInteger (0, 65535) << " " << uv->GetInteger (0, 65535) << " " << uv->GetInteger (0, 65535) << std::endl;
  seedFile << "GROUPING " << uv->GetInteger (0, 65535) << " " << uv->GetInteger (0, 65535) << " " << uv->GetInteger (0, 65535) << std::endl;
  seedFile << "ASSIGNMENT " << uv->GetInteger (0, 65535) << " " << uv->GetInteger (0, 65535) << " " << uv->GetInteger (0, 65535) << std::endl;
  seedFile << "BANDWIDTH " << uv->GetInteger (0, 65535) << " " << uv->GetInteger (0, 65535) << " " << uv->GetInteger (0, 65535) << std::endl; 
  seedFile.close ();

  brite::Brite br (confFile, "briteSeedFile.txt", "briteSeedFile.txt");

  m_topology = br.GetTopology ();
  BuildBriteNodeInfoList ();
  BuildBriteEdgeInfoList ();
  
  //remove temporary seed file
  remove("briteSeedFile.txt");
}

BriteTopologyHelper::~BriteTopologyHelper ()
{
  delete m_topology;

  while(!netDevices.empty())
  {
    delete netDevices.back();
    netDevices.pop_back();
  }
  
  while(!asLeafNodes.empty())
  {
    delete asLeafNodes.back();
    asLeafNodes.pop_back();
  }
}

void 
BriteTopologyHelper::BuildBriteNodeInfoList (void)
{
  brite::Graph *g = m_topology->GetGraph ();
  for (int i = 0; i < g->GetNumNodes (); ++i)
    {
      BriteNodeInfo nodeInfo;
      nodeInfo.nodeId = g->GetNodePtr (i)->GetId ();
      nodeInfo.xCoordinate = g->GetNodePtr (i)->GetNodeInfo ()->GetCoordX ();
      nodeInfo.yCoordinate = g->GetNodePtr (i)->GetNodeInfo ()->GetCoordY ();
      nodeInfo.inDegree = g->GetNodePtr (i)->GetInDegree ();
      nodeInfo.outDegree = g->GetNodePtr (i)->GetOutDegree ();

      switch(g->GetNodePtr(i)->GetNodeInfo()->GetNodeType()) 
        {
          case brite::NodeConf::RT_NODE:
 
            if(((brite::RouterNodeConf*)(g->GetNodePtr(i)->GetNodeInfo()))->GetASId() == -1)
              numAs = nodeInfo.asId = 0;
            else
              numAs = nodeInfo.asId = ((brite::RouterNodeConf*)(g->GetNodePtr(i)->GetNodeInfo()))->GetASId();
            
            switch (((brite::RouterNodeConf*)(g->GetNodePtr(i)->GetNodeInfo()))->GetRouterType()) 
              {
                case brite::RouterNodeConf::RT_NONE:
                  nodeInfo.type = "RT_NONE ";
                  break;
                case brite::RouterNodeConf::RT_LEAF:
                  nodeInfo.type = "RT_LEAF ";
                  break;
                case brite::RouterNodeConf::RT_BORDER:
                  nodeInfo.type = "RT_BORDER";
                  break;
                case brite::RouterNodeConf::RT_STUB:
                  nodeInfo.type = "RT_STUB ";
                  break;
                case brite::RouterNodeConf::RT_BACKBONE:
                  nodeInfo.type = "RT_BACKBONE ";
                  break;
                default:
                  NS_FATAL_ERROR ("Topology::Output(): Improperly classfied Router node encountered...");
              }
            break;
        
          case brite::NodeConf::AS_NODE:
            numAs = nodeInfo.asId =
              ((brite::ASNodeConf*)(g->GetNodePtr(i)->GetNodeInfo()))->GetASId();
            
            switch (((brite::ASNodeConf*)(g->GetNodePtr(i)->GetNodeInfo()))->GetASType()) 
              {
                case brite::ASNodeConf::AS_NONE:
                  nodeInfo.type = "AS_NONE ";
                  break;
                case brite::ASNodeConf::AS_LEAF:
                  nodeInfo.type = "AS_LEAF ";
                  break;
                case brite::ASNodeConf::AS_STUB:
                  nodeInfo.type = "AS_STUB ";
                  break;
                case brite::ASNodeConf::AS_BORDER:
                  nodeInfo.type = "AS_BORDER ";
                  break;
                case brite::ASNodeConf::AS_BACKBONE:
                  nodeInfo.type = "AS_BACKBONE ";
                  break;
                default:
                  NS_FATAL_ERROR ("Topology::Output(): Improperly classfied AS node encountered...");
              }
            break;
        }
      
      m_briteNodeInfoList.push_back (nodeInfo);
    }
    
    //Currently numAS stores the highest AS number.  We want numAs to store the number
    //of AS created in the topology.  Since AS numbering starts at 0 we add one to get
    //the correct count
    numAs++;
}

void 
BriteTopologyHelper::BuildBriteEdgeInfoList (void)
{
  brite::Graph *g = m_topology->GetGraph ();
  std::list<brite::Edge*>::iterator el;
  std::list<brite::Edge*> edgeList = g->GetEdges ();

  for (el = edgeList.begin (); el != edgeList.end (); el++)
    {
      BriteEdgeInfo edgeInfo;
      edgeInfo.edgeId = (*el)->GetId ();
      edgeInfo.srcId = (*el)->GetSrc ()->GetId ();
      edgeInfo.destId = (*el)->GetDst ()->GetId ();
      edgeInfo.length = (*el)->Length ();

      switch((*el)->GetConf()->GetEdgeType()) 
        {
          case brite::EdgeConf::RT_EDGE:
            edgeInfo.delay = ((brite::RouterEdgeConf*)((*el)->GetConf()))->GetDelay();
            edgeInfo.bandwidth = (*el)->GetConf()->GetBW();
            //If there is only one AS, BRITE will use -1 as AS Number.  We want it to be 0 instead.
            edgeInfo.asFrom = (((brite::RouterNodeConf*)((*el)->GetSrc()->GetNodeInfo()))->GetASId() == -1) ? 0 : ((brite::RouterNodeConf*)((*el)->GetSrc()->GetNodeInfo()))->GetASId();
            edgeInfo.asTo = (((brite::RouterNodeConf*)((*el)->GetDst()->GetNodeInfo()))->GetASId() == -1) ? 0 : ((brite::RouterNodeConf*)((*el)->GetDst()->GetNodeInfo()))->GetASId();
            break;

          case brite::EdgeConf::AS_EDGE:
            edgeInfo.delay =  -1;   /* No delay for AS Edges */
            edgeInfo.bandwidth = (*el)->GetConf()->GetBW();                       
            edgeInfo.asFrom = ((brite::ASNodeConf*)((*el)->GetSrc()->GetNodeInfo()))->GetASId();
            edgeInfo.asTo = ((brite::ASNodeConf*)((*el)->GetDst()->GetNodeInfo()))->GetASId();
            break;

          default:
            NS_FATAL_ERROR ("Topology::Output(): Invalid Edge type encountered...");
        }
    
      switch ((*el)->GetConf()->GetEdgeType()) 
        {
          case brite::EdgeConf::RT_EDGE: 
            switch (((brite::RouterEdgeConf*)(*el)->GetConf())->GetRouterEdgeType()) 
              {
                case brite::RouterEdgeConf::RT_NONE:
                  edgeInfo.type = "E_RT_NONE ";
                  break;
                case brite::RouterEdgeConf::RT_STUB:
                  edgeInfo.type = "E_RT_STUB ";
                  break;
                case brite::RouterEdgeConf::RT_BORDER:
                  edgeInfo.type = "E_RT_BORDER ";
                  break;
                case brite::RouterEdgeConf::RT_BACKBONE:
                  edgeInfo.type = "E_RT_BACKBONE ";
                  break;
                default:
                  NS_FATAL_ERROR ("Output(): Invalid router edge type...");
              }
          break;

          case brite::EdgeConf::AS_EDGE:
            switch (((brite::ASEdgeConf*)((*el)->GetConf()))->GetASEdgeType()) 
              {
                case brite::ASEdgeConf::AS_NONE:
                  edgeInfo.type = "E_AS_NONE ";
                  break;
                case brite::ASEdgeConf::AS_STUB:
                  edgeInfo.type = "E_AS_STUB ";
                  break;
                case brite::ASEdgeConf::AS_BORDER:
                  edgeInfo.type = "E_AS_BORDER ";
                  break;
                case brite::ASEdgeConf::AS_BACKBONE:
                  edgeInfo.type = "E_AS_BACKBONE ";
                  break;
                default:
                  NS_FATAL_ERROR ("BriteOutput(): Invalid AS edge type...");
              }
          break;
            
          default:
            NS_FATAL_ERROR ("BriteOutput(): Invalid edge type...");
            
        }

      m_briteEdgeInfoList.push_back (edgeInfo);
    }
}

NodeContainer&
BriteTopologyHelper::GetLeafNodesForAS (uint32_t asNum)
{
  return *asLeafNodes[asNum];
}

uint32_t 
BriteTopologyHelper::GetNNodesTopology () const
{
  return numNodes;
}

    
uint32_t 
BriteTopologyHelper::GetNEdgesTopology () const
{
  return numEdges;
}

uint32_t
BriteTopologyHelper::GetNAs (void) const
{
  return numAs;
}

uint32_t 
BriteTopologyHelper::GetSystemNumberForAs (uint32_t asNum) const 
{
  return systemForAs[asNum];
}

void
BriteTopologyHelper::BuildBriteTopology (const uint32_t systemCount)
{
  
  //determine as system number for each AS  
  for(uint32_t i = 0; i < numAs; ++i)
    {
      int val = i % systemCount;
      systemForAs.push_back(val);
      NS_LOG_INFO ("AS: " << i << " System: " << val);
    }
    
  //create nodes
  for(BriteTopologyHelper::BriteNodeInfoList::iterator it = m_briteNodeInfoList.begin(); it != m_briteNodeInfoList.end(); ++it)
    {
        nodes.Add(CreateObject<Node> (GetSystemNumberForAs ((*it).asId)));    
        numNodes++;
    }

  ConstructTopology ();  
}

void 
BriteTopologyHelper::InstallStack (InternetStackHelper& stack)
{
  stack.Install (nodes);
}

void 
BriteTopologyHelper::AssignIpv4Addresses (Ipv4AddressHelper& address)
{
  //assign IPs to each AS
  for(unsigned int i = 0; i < netDevices.size(); ++i)
    {
      address.Assign(*netDevices[i]);
      address.NewNetwork();  
    }
}

void 
BriteTopologyHelper::AssignIpv6Addresses (Ipv6Address network, Ipv6Prefix prefix)
{
  Ipv6AddressGenerator::Init (network, prefix);
  Ipv6Address v6network;
  Ipv6AddressHelper addressHelper;

  //assign IPs to each AS
  for(unsigned int i = 0; i < netDevices.size (); ++i)
    {
      v6network = Ipv6AddressGenerator::GetNetwork (prefix);
      addressHelper.SetBase (v6network, prefix);
      addressHelper.Assign (*netDevices[i]);
      Ipv6AddressGenerator::NextNetwork (prefix);
    }
}

void
BriteTopologyHelper::ConstructTopology ()
{
  //create one device container for each AS and one node container to hold leaf nodes for attaching
  for(uint32_t i = 0; i < numAs; ++i)
    {
      netDevices.push_back(new NetDeviceContainer());
      asLeafNodes.push_back(new NodeContainer());
    }

  for (BriteTopologyHelper::BriteEdgeInfoList::iterator it = m_briteEdgeInfoList.begin(); it != m_briteEdgeInfoList.end(); ++it) 
    {
      // Set the link delay
      // The brite value for delay is given in milliseconds
      m_britePointToPointHelper.SetChannelAttribute ("Delay", 
                               TimeValue (MilliSeconds ((*it).delay)));

      // The brite value for data rate is given in Mbps                        
      m_britePointToPointHelper.SetDeviceAttribute ("DataRate", 
                              DataRateValue (DataRate ((*it).bandwidth * mbpsToBps)));

      netDevices[(*it).asFrom]->Add (m_britePointToPointHelper.Install (nodes.Get ((*it).srcId), 
                            nodes.Get ((*it).destId)));
      
      numEdges++;
    }
  
  //iterate through all nodes and add leaf nodes for each AS 
  for(BriteTopologyHelper::BriteNodeInfoList::iterator it = m_briteNodeInfoList.begin(); it != m_briteNodeInfoList.end(); ++it)
    {
      if((*it).type == "RT_LEAF ")
      {
        asLeafNodes[(*it).asId]->Add (nodes.Get ((*it).nodeId));
      }
    }     
}

PointToPointHelper&
BriteTopologyHelper::GetPointToPointHelperForTopology ()
{
  return m_britePointToPointHelper;
}

void
BriteTopologyHelper::BuildBriteTopology ()
{

  //not using MPI so each AS is on system number 0
  for(uint32_t i = 0; i < numAs; ++i)
    {
      systemForAs.push_back(0);
    }

  //create all nodes with system number 0
  nodes.Create (m_briteNodeInfoList.size());

  numNodes = m_briteNodeInfoList.size();

  ConstructTopology ();

}

} // namespace ns3
